def get_config( default=False ):
    """
    Get configuration values as a list of strings which can be logged, stored, or evaluated.

    The default values (returned when default is True) are the configuration values after all config files have been evaluated but before the path values have been expanded using os.path.expanduser and os.path.abspath. Modules that use the command line to change config values may also not update the default values. User actions in a CASA session will also typically not change the default values.

    Parameters
       default (bool=False) - If True, return the default values.

    Returns
       list[str] - list of configuration strings
    """

    from .. import config as _config
    if not default :
        valsObj = _config
    else:
        valsObj = _config._config_defaults

    return list( map( lambda name: f'{name} = {repr(getattr(valsObj,name))}', _config.__defaults ) )



# Copyright 2023 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
this module will be included in the api
"""

def get_data_info(path=None, logger=None):
    """
    Get the summary information on the 3 types of data managed by casaconfig.

    The return value is a dictionary by type of of data : 'casarundata', 'measures',
    and 'release'.

    The path is the location to use to search for the installed release information.
    The path argument defaults to config.measurespath when not set.

    The casarundata and measures type value are each a dictionary of 'version' 
    and 'date' where version is the version string and date is the date when it  
    was installed These values are taken from the readme.txt file for each type.

    The 'release' type is from the release_data_readme.txt file which is copied
    into place when a modular CASA is built. It consists of a dictionary of
    'casarundata' and 'measures' where the values are the version strings for each
    type of data for this release. That allows casaconfig to install that
    casarundata version for testing purposes. The release information does not depend 
    on the path argument since it is found in the casaconfig module.

    If path is empty or does not exist then the return value for the 'casarundata' and
    'measures' types is None.

    If no readme.txt file can be found at path but the contents of path have the directories
    expected for casarundata then the version returned for 'casarundata' is 'unknown' and
    the date is an empty string. The path location may contain casarundata from a legacy
    installation of CASA data. CASA may be able to use the files at this location but they
    can not be maintained by casaconfig. If the path is not empty but does not appear to 
    contain legacy casarundata then the version is 'invalid'.

    If no readme.txt file can be found for the measures at path/geodetic but both the geodetic
    and ephemeris directories are present in path then the version returned for 'measures' is
    'unknown' and the date is an empty string. The path location may contain measures data from
    a legacy installation of CASA data. CASA may be able to use any measures tables at this
    location by they can not be maintained by casaconfig. If the path is not empty but does not
    appear to contain legacy measures data then the version is 'invalid'.

    If no casadata release information is found or the contents are unexpected the returned
    value for 'release' is None and the "--reference-testing" option will not do anything
    for this installation of casaconfig. This will be the case for a modular installation.

    If path has not been set (has a value of None) then the returned value will be None. This
    likely means that a casasiteconfig.py exists but has not yet been edited to set measurespath.

    Parameters
       - path (str) - Folder path to find the casarundata and measures data information. If not set then config.measurespath is used.
       - logger (casatools.logsink=None) - Instance of the casalogger to use for writing messages. Default None writes messages to the terminal.
    
    Returns
       dict - a dictionary by type, 'casarundata', 'measures', 'release' where each type is a dictionary containing 'version' and 'date'. A return value of None indicates path is unset. A value of None for that type means no information could be found about that type.
    """

    # when None is returned, path wasn't set
    result = None

    import os
    import importlib.resources
    from .print_log_messages import print_log_messages
    from .read_readme import read_readme

    if path is None:
        from .. import config as _config
        path = _config.measurespath

    if path is None:
        # it's not being set in a config file, probably casasiteconfig.py is being used but has not been edited
        print_log_messages('path is None and has not been set in config.meausrespath (probably casasiteconfig.py). Provide a valid path and retry.', logger, True)

        return None
    
    path = os.path.abspath(os.path.expanduser(path))

    result = {'casarundata':None, 'measures':None, 'release':None}

    # casarundata and measures 

    if os.path.isdir(path) and (len(os.listdir(path))>0):
        # there's something at path, look for the casarundata readme
        datareadme_path = os.path.join(path,'readme.txt')
        if os.path.exists(datareadme_path):
            # the readme exists, get the info
            result['casarundata'] = {'version':'error', 'date':''}
            readmeContents = read_readme(datareadme_path)
            if readmeContents is not None:
                currentVersion = readmeContents['version']
                currentDate = readmeContents['date']
                # this one is just a check that the manifest at least exists
                if len(readmeContents['extra']) > 0:
                    result['casarundata'] = {'version':currentVersion, 'date':currentDate}
        else:
            # does it look like it's probably casarundata?
            expected_dirs = ['alma','catalogs','demo','ephemerides','geodetic','gui','nrao']
            ok = True
            for d in expected_dirs:
                if not os.path.isdir(os.path.join(path,d)): ok = False
            if ok:
                # probably casarundata
                result['casarundata'] = {'version':'unknown', 'date':''}
            else:
                # probably not casarundata
                result['casarundata'] = {'version':'invalid', 'date':''}

        # look for the measures readme
        measuresreadme_path = os.path.join(path,'geodetic/readme.txt')
        if os.path.exists(measuresreadme_path):
            # the readme exists, get the info
            result['measures'] = {'version':'error', 'date':''}
            readmeContents = read_readme(measuresreadme_path)
            if readmeContents is not None:
                currentVersion = readmeContents['version']
                currentDate = readmeContents['date']
                result['measures'] = {'version':currentVersion,'date':currentDate}
        else:
            # does it look like it's probably measuresdata?
            # path should have ephemerides and geodetic directories
            if os.path.isdir(os.path.join(path,'ephemerides')) and os.path.isdir(os.path.join(path,'geodetic')):
                result['measures'] = {'version':'unknown', 'date':''}
            else:
                # probably not measuresdata
                result['measures'] = {'version':'invalid', 'date':''}

    # release data versions
    if importlib.resources.is_resource('casaconfig','release_data_readme.txt'):
        try:
            casarundataVersion = None
            measuresVersion = None
            ok = True
            reason = None
            readme_lines = importlib.resources.read_text('casaconfig','release_data_readme.txt').split('\n')
            for readmeLine in readme_lines:
                # lines must contain something and not start with #
                if len(readmeLine) > 0 and readmeLine[0] != '#':
                    splitLine = readmeLine.split(':')
                    if len(splitLine) == 2:
                        lineType = splitLine[0].strip()
                        lineVers = splitLine[1].strip()
                        if lineType == 'casarundata':
                            if casarundataVersion is not None:
                                ok = False
                                reason = "duplicate casarundata lines"
                                break
                            casarundataVersion = lineVers
                        elif lineType == 'measures':
                            if measuresVersion is not None:
                                ok = False
                                reason = "duplicate measures lins"
                                break
                            measuresVersion = lineVers
                        else:
                            ok = False
                            reason = "Unexpected type : %s" % lineType
                            break
                    else:
                        ok = False
                        reason = "Missing or too many ':' separators"
            if (casarundataVersion is None or measuresVersion is None) and ok:
                ok = False
                reason = "missing one or more version strings for expected casarundata and measures types"

            if not ok:
                print_log_messages("Incorrectly formatted release_data_readme.txt. %s" % reason, logger, True)
                # leave 'release' as None
            else:
                result['release'] = {'casarundata':casarundataVersion, 'measures':measuresVersion}
        except:
            print("Unexpected error reading release_data_readme.txt")
            # leave 'release' as None
    else:
        # no release information available, probably a modular install only
        # leave 'release' as None
        pass

    return result



# Copyright 2020 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
this module will be included in the api
"""

def measures_update(path=None, version=None, force=False, logger=None, auto_update_rules=False):
    """
    Retrieve IERS data used for measures calculations from ASTRON server
    
    Original data source is here: https://www.iers.org/IERS/EN/DataProducts/data.html

    CASA maintains a separate Observatories table which is available in the casarundata
    collection through pull_data and data_update. The Observatories table found at ASTRON
    is not installed by measures_update and any Observatories file at path will not be changed
    by using this function.

    A text file (readme.txt in the geodetic directory in path) records the version string
    and the date when that version was installed in path.

    If path is None then config.measurespath is used.

    If the version requested matches the one in that text file then this function does
    nothing unless force is True.

    If a specific version is not requested (the default) and the date in that text file
    is today, then this function does nothing unless force is True even if there is a more
    recent version available from the ASTRON server.

    When auto_update_rules is True then path must be owned by the user, force must be False,
    and the version must be None. This is used during casatools initialization when 
    measures_auto_update is True. Automatic updating happens during casatools initialization
    so that the updated measures are in place before any tool needs to use them.

    Using measures_update after casatools has started should always be followed by exiting 
    and restarting casa (or the casatools module if modular casa components are being used).

    A file lock is used to prevent more that one data update (pull_data, measures_update,
    or data_update) from updating any files in path at the same time. When locked, the 
    lock file (data_update.lock in path) contains information about the process that
    has the lock. When measures_update gets the lock it checks the readme.txt file in path
    to make sure that an update is still necessary (if force is True then an update 
    always happens). If the lock file is not empty then a previous update of path (pull_data,
    data_update, or measures_update) did not exit as expected and the contents of path are
    suspect. In that case, an error will be reported and nothing will be updated. The lock
    file can be checked to see the details of when that file was locked. The lock file can be
    removed and measures_update can be tried again. It may be safest in that case to remove path
    completely or use a different path and use pull_data to install a fresh copy of the
    desired version.

    Care should be used when using measures_update outside of the normal automatic
    update that other casa sessions are not using the same measures at the same time,
    especially if they may also be starting at that time. If a specific version is
    requested or force is True there is a risk that the measures may be updated while
    one of those other sessions are trying to load the same measures data, leading to
    unpredictable results. The lock file will prevent simulateous updates from
    happening but if each simultanous update eventually updates the same measures
    location (because force is True or the updates are requesting different versions)
    then the measures that any of those simultanous casatools modules sees is 
    unpredictable. Avoid multiple, simultanous updates outside of the automatic
    update process.

    **Note:** measures_update requires that the expected readme.txt file already exists in
    the geodetic directory at path. If that file does not exist or can not be interpreted as
    expected then measures_update will return without updating any data.

    **Note:** if auto_update_rules is True the user must own path (in addition to having 
    read and write permissions there). The version must then also be None and the force option 
    must be False.


    Parameters
       - path (str=None) - Folder path to place updated measures data. Must contain a valid geodetic/readme.txt. If not set then config.measurespath is used.
       - version (str=None) - Version of measures data to retrieve (usually in the form of yyyymmdd-160001.ztar, see measures_available()). Default None retrieves the latest.
       - force (bool=False) - If True, always re-download the measures data. Default False will not download measures data if already updated today unless the version parameter is specified and different from what was last downloaded.
       - logger (casatools.logsink=None) - Instance of the casalogger to use for writing messages. Default None writes messages to the terminal
       - auto_update_rules (bool=False) - If True then the user must be the owner of path, version must be None, and force must be False.
        
    Returns
       None
    
    """
    import os
    import pkg_resources
    from datetime import datetime
    import sys

    from ftplib import FTP
    import tarfile
    import re
    import ssl
    import urllib.request
    import certifi
    import fcntl

    from .print_log_messages import print_log_messages
    from .get_data_lock import get_data_lock
    from .read_readme import read_readme
    
    if path is None:
        from .. import config as _config
        path = _config.measurespath

    if path is None:
        # it's not being set in a config file, probably casasiteconfig.py is being use but has not been edited
        print_log_messages('path is None and has not been set in config.measurespath (probably casasiteconfig.py). Provide a valid path and retry.', logger, True)
        return

    path = os.path.expanduser(path)

    if auto_update_rules:
        if version is not None:
            print_log_messages('auto_update_rules requires that version be None', logger, True)
            return
        if force:
            print_log_messages('force must be False when auto_update_rules is True', logger, True)
            return
        if (not os.path.isdir(path)) or (os.stat(path).st_uid != os.getuid()):
            print_log_messages('path must exist as a directory and it must be owned by the user when auto_update_rules is True', logger, True)
            return
    
    if not os.path.exists(path):
        # make dirs all the way down, if possible
        os.makedirs(path)
        
    current = None
    updated = None
    today_string = datetime.today().strftime('%Y-%m-%d')

    # first, does this look like it needs to be updated

    # if measures are already preset, get their version
    readme_path = os.path.join(path,'geodetic/readme.txt')
    if os.path.exists(readme_path):
        readmeInfo = read_readme(readme_path)
        if readmeInfo is not None:
            current = readmeInfo['version']
            updated = readmeInfo['date']

    # don't re-download the same data
    if not force:
        if ((version is not None) and (version == current)) or ((version is None) and (updated == today_string)):
            print_log_messages('measures_update current measures detected in %s, using version %s' % (path, current), logger)
            return

    # path must be writable with execute bit set
    if (not os.access(path, os.W_OK | os.X_OK)) :
        print_log_messages('No permission to write to the measures path, cannot update : %s' % path, logger, True)
        return

    # an update needs to happen

    # lock the measures_update.lock file
    lock_fd = None
    try:
        print_log_messages('measures_update measures need to be updated, acquiring the lock ... ', logger)

        lock_fd = get_data_lock(path, 'measures_update')
        # if lock_fd is None it means the lock file was not empty - because we know that path exists at this point
        if lock_fd is None:
            print_log_messages('The lock file at %s is not empty.' % path, logger, True)
            print_log_messages('A previous attempt to update path may have failed or exited prematurely.', logger, True)
            print_log_messages('Remove the lock file and set force to True with the desired version (default to most recent).', logger, True)
            print_log_messages('It may be best to completely repopulate path using pull_data and measures_update.', logger, True)
            return

        do_update = force

        if not do_update:
            # recheck the readme file, another update may have already happened before the lock was obtained
            current = None
            updated = None
            if os.path.exists(readme_path):
                readmeInfo = read_readme(readme_path)
                if readmeInfo is not None:
                    current = readmeInfo['version']
                    updated = readmeInfo['date']

            if ((version is not None) and (version == current)) or ((version is None) and (updated == today_string)):
                # no update will be done, version is as requested or it's already been updated today
                print_log_messages('measures_update current measures detected in %s, using version %s' % (path, current), logger)
            else:
                # an update is needed
                do_update = True

        if do_update:
            if force:
                print_log_messages('meaures_update a measures update has been requested by the force argument', logger)

            print_log_messages('measures_update connecting to ftp.astron.nl ...', logger)

            ftp = FTP('ftp.astron.nl')
            rc = ftp.login()
            rc = ftp.cwd('outgoing/Measures')
            files = sorted([ff for ff in ftp.nlst() if (len(ff) > 0) and (not ff.endswith('.dat')) and (ftp.size(ff) > 0)])

            # target filename to download
            target = files[-1] if version is None else version
            if target not in files:
                print_log_messages('measures_update cant find specified version %s' % target, logger, True)

            else:
                # there are files to extract, remove the readme.txt file in case this dies unexpectedly
                if os.path.exists(readme_path):
                    os.remove(readme_path)
    
                with open(os.path.join(path,'measures.ztar'), 'wb') as fid:
                    print_log_messages('measures_update downloading %s from ASTRON server to %s ...' % (target, path), logger)
                    ftp.retrbinary('RETR ' + target, fid.write)

                ftp.close()
    
                # extract from the fetched tarfile
                with tarfile.open(os.path.join(path,'measures.ztar'),mode='r:gz') as ztar:
                    # the list of members to extract
                    x_list = []
                    for m in ztar.getmembers() :
                        # always exclude the Observatories table and  *.old names in geodetic
                        if not((re.search('geodetic',m.name) and re.search('.old',m.name)) or re.search('Observatories',m.name)):
                            x_list.append(m)

                    # use the 'data' filter if available, revert to previous 'fully_trusted' behavior of not available
                    ztar.extraction_filter = getattr(tarfile, 'data_filter', (lambda member, path: member))
                    ztar.extractall(path=path,members=x_list)
                    ztar.close()

                os.system("rm %s" % os.path.join(path, 'measures.ztar'))

                # update the readme.txt file
                with open(readme_path,'w') as fid:
                    fid.write("# measures data populated by casaconfig\nversion : %s\ndate : %s" % (target, datetime.today().strftime('%Y-%m-%d')))

                print_log_messages('measures_update updated measures data at %s' % path, logger)
            
            # closing out the do_update

        # closing out the try block
        # truncate the lock file
        lock_fd.truncate(0)
        
    except Exception as exc:
        print_log_messages("ERROR! : Unexpected exception while updating measures at %s" % path, logger, True)
        print_log_messages("ERROR! : %s" % exc, logger, True)
        # leave the contents of the lock file as is to aid in debugging
        
    # if the lock file is not closed, do that now to release the lock
    if lock_fd is not None and not lock_fd.closed:
        lock_fd.close()

    return



# Copyright 2023 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
this module will be included in the api
"""

def data_update(path=None, version=None, force=False, logger=None, auto_update_rules=False):
    """
    Check for updates to the installed casarundata and install the update or change to 
    the requested version when appropriate.
    
    A text file (readme.txt at path) records the version string, the date 
    when that version was installed in path, and the files installed into path. That file
    must already exist in path in order to use this function. Use pull_data to install
    casarundata into a new location.

    If path is None then config.measurespath is used.

    If the version is None (the default) then the most recent version returned by 
    data_available is used.

    If version is "release" then the version associated with the release in 
    the dictionary returned by get_data_info is used. If there is no release
    version information known then an error message is printed and nothing is
    updated.

    If the version requested matches the one in the readme.txt file at path then this function does
    nothing unless force is True.

    If a specific version is not requested (the default) and the date in the readme.txt file
    at path is today, then this function does nothing unless force is True even if there is a more
    recent version available from the CASA server.

    When auto_update_rules is True then path must be owned by the user, force must be
    False and the version must be None. This is used during casatools initialization when
    data_auto_update is True. Automatic updating happens during casatools initialization
    so that the updated casarundata and measures are in place before any tool needs to use them.

    If an update is to be installed the previously installed files, as listed in the 
    readme.txt file at path, are removed before the contents of the version 
    being installed are unpacked. This includes the set of measures tables
    that are part of that casarundata version. A data update is typically followed by a
    measures_update to ensure that the most recent measures data are installed.

    A file lock is used to prevent more that one data update (pull_data, measures_update,
    or data_update) from updating any files in path at the same time. When locked, the 
    lock file (data_update.lock in path) contains information about the process that
    has the lock. When data_update gets the lock it checks the readme.txt file in path
    to make sure that an update is still necessary (if force is True then an update 
    always happens). If the lock file is not empty then a previous update of path (pull_data,
    data_update, or measures_update) did not exit as expected and the contents of path are
    suspect. In that case, an error will be reported and nothing will be updated. The lock
    file can be checked to see the details of when that file was locked. The lock file can be
    removed and data_update can be tried again. It may be safest in that case to remove path
    completely or use a different path and use pull_data to install a fresh copy of the
    desired version.

    Some of the tables installed by data_update are only read when casatools starts. Use of 
    data_update should typically be followed by a restart of CASA so that any changes are seen by
    the tools and task that use this data.

    **Note:** data_update requires that the expected readme.txt file already exists at the top-level
    directory at path. If the file does not exist or can not be interpreted as expected then
    data_update will return without updating any data.

    **Note:** if auto_update_rules is True the user must own path (in addition to having read and
    write permissions there). The version must then also be None and the force option must be False.

    **Note:** the most recent casarundata may not include the most recent measures data. A data_update
    is typically followed by a measures_update.

    Parameters
       - path (str=None) - Folder path to update. Must contain a valid readme.txt. If not set then config.measurespath is used.
       - version (str=None) - Version of casarundata to retrieve (usually in the form of casarundata-x.y.z.tar.gz, see data_available()). Default None retrieves the latest.
       - force (bool=False) - If True, always re-download the casarundata. Default False will not download casarundata if already updated today unless the version parameter is specified and different from what was last downloaded.
       - logger (casatools.logsink=None) - Instance of the casalogger to use for writing messages. Default None writes messages to the terminal.
       - auto_update_rules (bool=False) - If True then the user must be the owner of path, version must be None, and force must be False.
        
    Returns
       None
    
    """

    import os
    from datetime import datetime

    from .data_available import data_available
    from .print_log_messages import print_log_messages
    from .get_data_lock import get_data_lock
    from .do_pull_data import do_pull_data
    from .get_data_info import get_data_info
    from .read_readme import read_readme

    if path is None:
        from .. import config as _config
        path = _config.measurespath

    if path is None:
        # it's not being set in a config file, probably casasiteconfig.py is being use but has not been edited
        print_log_messages('path is None and has not been set in config.measurespath (probably casasiteconfig.py). Provide a valid path and retry.', logger, True)
        return

    path = os.path.expanduser(path)
    readme_path = os.path.join(path, 'readme.txt')

    if auto_update_rules:
        if version is not None:
            print_log_messages('auto_update_rules requires that version be None', logger, True)
            return
        if force:
            print_log_messages('force must be False when auto_update_rules is True', logger, True)
            return
        if (not os.path.isdir(path)) or (os.stat(path).st_uid != os.getuid()):
            print_log_messages('path must exist as a directory and it must be owned by the user when auto_update_rules is True', logger, True)
            return        

    # readme must exist
    if not os.path.exists(readme_path):
        print_log_messages('No readme.txt file found at path. Nothing updated or checked.', logger, True);
        return

    # path must be writable with execute bit set
    if (not os.access(path, os.W_OK | os.X_OK)) :
        print('No permission to write to path, cannot update : %s' % path, file=sys.stdout)
        if logger is not None: logger.post('No permission to write to path, cannot update : %s' % path, 'ERROR')
        return
    
    # try and digest the readme file

    installed_files = []
    currentVersion = []
    currentDate = []
    dataReadmeInfo = read_readme(readme_path)
    if dataReadmeInfo is not None:
        currentVersion = dataReadmeInfo['version']
        currentDate = dataReadmeInfo['date']
        installed_files = dataReadmeInfo['extra']
    else:
        print_log_messages('The readme.txt file at path could not be read as expected', logger, True)
        print_log_messages('choose a different path or empty this  path and try again using pull_data', logger, True)
        # no lock has been set yet, safe to simply return here
        return

    if (len(installed_files) == 0):
        # this shouldn't happen
        print_log_messages('The readme.txt file at path did not contain the expected list of installed files', logger, True)
        print_log_messages('choose a different path or empty this path and try again using pull_data', logger, True)
        # no lock has been set yet, safe to simply return here
        return

    today_string = datetime.today().strftime('%Y-%m-%d')
    if version is None and force is False and currentDate == today_string:
        # if version is None, currentDate is today and force is False then return without checking for any newer versions
        print_log_messages('data_update current casarundata detected in %s, using version %s' % (path, currentVersion), logger)
        # no lock has been set yet, safe to simply return here
        return

    available_data = data_available()
    requestedVersion = version

    if requestedVersion is None:
        requestedVersion = available_data[-1]

    expectedMeasuresVersion = None
    if requestedVersion == 'release':
        # use the release version from get_data_info
        releaseInfo = get_data_info()['release']
        if releaseInfo is None:
            print_log_messages('No release info found, pull_data can not continue', logger, True)
            return
        requestedVersion = releaseInfo['casarundata']
        expectedMeasuresVersion = releaseInfo['measures']

    if requestedVersion not in available_data:
        print_log_messages('Requested casarundata version %s was not found. See available_data for a list of available casarundata versions.' % requestedVersion)
        # no lock has been set yet, safe to simply return here
        return

    # don't update if force is false and the requested version is already installed
    if force is False and (currentVersion == requestedVersion):
        if expectedMeasuresVersion is not None:
            # the 'release' version has been requested, need to check the measures version
            # assume a force is necessary until the measures version is known to be OK
            force = True
            measures_readme_path = os.path.join(path,'geodetic/readme.txt')
            if os.path.exists(measures_readme_path):
                measuresReadmeInfo = read_readme(measures_readme_path)
                if measuresReadmeInfo is not None:
                    measuresVersion = measuresReadmeInfo['version']
                    if measuresVersion == expectedMeasuresVersion:
                        # it's OK, do not force
                        force = False
                # if measuresReadmeInfo is None then that's a problem and force remains True
            if not force:
                print_log_messages('data_update requested "release" version of casarundata and measures are already installed.', logger)
                # no lock has been set yet, safe to simply return here
                return
        else:
            # normal usage, ok to return now
            print_log_messages('Requested casarundata is installed in %s, using version %s' % (path, currentVersion), logger)
            # no lock has been set yet, safe to simply return here
            return

    # an update appears necessary

    lock_fd = None
    try:
        print_log_messages('data_update using version %s, acquiring the lock ... ' % requestedVersion, logger)

        lock_fd = get_data_lock(path, 'data_update')
        # if lock_fd is None it means the lock file was not empty - because we know that path exists at this point
        if lock_fd is None:
            print_log_messages('The lock file at %s is not empty.' % path, logger, True)
            print_log_messages('A previous attempt to update path may have failed or exited prematurely.', logger, True)
            print_log_messages('Remove the lock file and set force to True with the desired version (default to most recent).', logger, True)
            print_log_messages('It may be best to completely repopulate path using pull_data and measures_update.', logger, True)
            return

        do_update = True
        # it's possible that another process had path locked and updated the readme with new information, re-read it
        dataReadmeInfo = read_readme(readme_path)
        if dataReadmeInfo is not None:
            currentVersion = dataReadmeInfo['version']
            currentDate = dataReadmeInfo['date']
            installedFiles = dataReadmeInfo['extra']
            if ((currentVersion == requestedVersion) and (not force)):
                if expectedMeasuresVersion is not None:
                    # this is a 'release' update request, need to check that the measures version is also now OK
                    measures_readme_path = os.path.update(path,'geodetic/readme.txt')
                    if os.path.exists(measures_readme_path):
                        measuresReadmeInfo = read_readme(measures_readme_path)
                        if measuresReadmeInfo is not None:
                            measuresVersion = measuresReadmeInfo['version']
                            if measuresVersion == expectedMeasuresVersion:
                                do_update = False
                        # if measuresReadmeInfo is None there was a problem which requires a full update so do_update remains True
                    if not do_update:
                        print_log_messages('data update requested "release" version of casarundata and measures are already installed.', logger)
                else:
                    # nothing to do here, already at the expected version and an update is not being forced
                    do_update = False
                    print_log_messages('data_update requested version is already installed.', logger)
                    
            if do_update:
                # update is still on, check the manifest
                if len(installed_files) == 0:
                    # this shouldn't happen, do not do an update
                    do_update = False
                    print_log_messages('The readme.txt file read at path did not contain the expected list of installed files', logger, True)
                    print_log_messages('This should not happen unless multiple sessions are trying to update data at the same time and one experienced problems or was done out of sequence', logger, True)
                    print_log_messages('Check for other updates in process or choose a different path or clear out this path and try again using pull_data or update_all', logger, True)
        else:
            # this shouldn't happen, do not do an update
            do_update = False
            print_log_messages('Unexpected problem reading readme.txt file during data_update, can not safely update to the requested version', logger, True)
            print_log_messages('This should not happen unless multiple sessions are trying to update at the same time and one experienced problems or was done out of sequence', logger, True)
            print_log_messages('Check for other updates in process or choose a different path or clear out this path and try again using pull_data or update_all', logger, True)

    
        if do_update:
            do_pull_data(path, requestedVersion, installed_files, currentVersion, currentDate, logger)

        # truncate the lock file
        lock_fd.truncate(0)

    except Exception as exc:
        print_log_messages('ERROR! : Unexpected exception while populating casarundata version %s to %s' % (requestedVersion, path), logger, True)
        print_log_messages('ERROR! : %s' % exc, logger, True)
        # leave the contents of the lock file as is to aid in debugging
        # import traceback
        # traceback.print_exc()

    # if the lock file is not closed, do that now to release the lock
    if lock_fd is not None and not lock_fd.closed:
        lock_fd.close()
    
    return



# Copyright 2020 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
this module will be included in the api
"""

def pull_data(path=None, version=None, force=False, logger=None):
    """
    Pull the casarundata contents from the CASA host and install it in path.

    The path must either contain a previously installed version of casarundata 
    or it must not exist or be empty.

    If path is None then config.measurespath will be used.

    If version is None (the default) then the most recent version is pulled.

    If version is "release" then the version associated with the release in 
    the dictionary returned by get_data_info is used. If there is no release
    version information known then an error message is printed and nothing is
    updated.

    A text file (readme.txt at path) records the version string, the date
    when that version was installed in path, and the files installed into path.
    This file is used to determine if the contents are a previously installed 
    version.

    If the version to be pulled matches the version in the readme.txt file then this 
    function does nothing unless force is true. No error messages will result when the
    version already matches what was previously installed (no installation is then
    necessary unless force is True).

    The measures tables included in casarundata will typically 
    not be the most recent version. To get the most recent measures data, measures_update
    should be used after pull_data.

    If path contains a previously installed version then all of the files listed in 
    the manifest part of the readme.txt file are first removed from path.

    A file lock is used to prevent more than one data update (pull_data, measures_update,
    or data_update) from updating any files in path at the same time. When locked, the
    lock file (data_update.lock in path) contains information about the process that
    has the lock. When pull_data gets the lock it checks the readme.txt file in path
    to make sure that a copy of the data should be pulled (the version doesn't match what
    was requested). If force is True an update always happens. If the lock file is not
    empty then a previous update of path (pull_data, data_update, or measures_update)
    did not exit as expected and the contents of path are suspect. In that case, pull_data
    will report that as an error and nothing will be updated. The lock file can be checked
    to see the details of when that file was locked. The lock file can be removed and
    pull_data can then be used to install the desired version. It may be safest in that case
    to remove path completely or use a different path and run pull_data to install 
    a fresh copy of the desired version.

    Some of the tables installed by pull_data are only read when casatools starts. Use of 
    pull_data should typically be followed of CASA by a restart so that 
    any changes are seen by the tools and tasks that use this data.

    Parameters
       - path (str) - Folder path to place casarundata contents. It must be empty or not exist or contain a valid, previously installed version. If not set then config.measurespath is used.
       - version (str=None) - casadata version to retrieve. Default None gets the most recent version.
       - force (bool=False) - If True, re-download the data even when the requested version matches what is already installed. Default False will not download data if the installed version matches the requested version.
       - logger (casatools.logsink=None) - Instance of the casalogger to use for writing messages. Messages are always written to the terminal. Default None does not write any messages to a logger.

    Returns
       None

    """

    import os

    from .data_available import data_available
    from .print_log_messages import print_log_messages
    from .get_data_lock import get_data_lock
    from .do_pull_data import do_pull_data
    from .get_data_info import get_data_info
    from .read_readme import read_readme

    if path is None:
        from .. import config as _config
        path = _config.measurespath

    if path is None:
        # it's not being set in a config file, probably casasiteconfig.py is being use but has not been edited
        print_log_messages('path is None and has not been set in config.measurespath (probably casasiteconfig.py). Provide a valid path and retry.', logger, True)
        return

    path = os.path.expanduser(path)
    readme_path = os.path.join(path, 'readme.txt')

    installed_files = []
    available_data = None
    currentVersion = None
    currentDate = None
    
    # attempt a pull if path does not exist or is empty:
    do_pull = (not os.path.exists(path)) or (len(os.listdir(path))==0)
    if not do_pull:
        # if the readme_path exists, find the current version, install date, and installed files
        if os.path.exists(readme_path):
            readmeInfo = read_readme(readme_path)
            if readmeInfo is not None:
                currentVersion = readmeInfo['version']
                currentDate = readmeInfo['date']
                installed_files = readmeInfo['extra']
            else:
                print_log_messages('destination path is not empty and the readme.txt file found there could not be read as expected', logger, True)
                print_log_messages('choose a different path or empty this path and try again', logger, True)
                # no lock has been set yet, safe to simply return here
                return

            if (len(installed_files) == 0):
                # this shouldn't happen
                print_log_messages('destination path is not empty and the readme.txt file found there did not contain the expected list of installed files', logger, True)
                print_log_messages('choose a different path or empty this path and try again', logger, True)
                # no lock as been set yet, safe to simply return here
                return
            
            # the readme file looks as expected, pull if the version is different or force is true
            if version is None:
                # use most recent available
                available_data = data_available()
                version = available_data[-1]
            do_pull = (version!=currentVersion) or force

            if not do_pull:
                # it's already at the expected version and force is False, nothing to do
                # safe to return here, no lock has been set
                return
        else:
            # path exists and is not empty
            print_log_messages('destination path exists, is not empty, and does not contain the expected readme.txt file', logger, True)
            print_log_messages('choose a different path or empty this path and try again', logger, True)
            # no lock has been set yet, safe to simply return here
            return

    # a pull will happen, unless the version string is not available

    if available_data is None:
        available_data = data_available()

    if version is None:
        # need a version, use most recent available
        version = available_data[-1]

    expectedMeasuresVersion = None
    if version == 'release':
        # use the release version from get_data_info
        releaseInfo = get_data_info()['release']
        if releaseInfo is None:
            print_log_messages('No release info found, pull_data can not continue', logger, True)
            return
        version = releaseInfo['casarundata']
        expectedMeasuresVersion = releaseInfo['measures']

    if version not in available_data:
        print_log_messages('version %s not found on CASA server, use casaconfig.data_available for a list of casarundata versions' % version, logger, True)
        return

    if not os.path.exists(path):
        # make dirs all the way down path if possible
        os.makedirs(path)

    # lock the data_update.lock file
    lock_fd = None
    try:
        print_log_messages('pull_data using version %s, acquiring the lock ... ' % version, logger)

        lock_fd = get_data_lock(path, 'pull_data')
        # if lock_fd is None it means the lock file was not empty - because we know that path exists at this point
        if lock_fd is None:
            print_log_messages('The lock file at %s is not empty.' % path, logger, True)
            print_log_messages('A previous attempt to update path may have failed or exited prematurely.', logger, True)
            print_log_messages('It may be best to do a fresh pull_data on that location.', logger, True)
            print_log_messages('Remove the lock file and set force to True with the desired version (default to the most recent).', logger, True)
            return
        
        do_pull = True
        if not force and os.path.exists(readme_path):
            # the readme file needs to be reread here
            # it's possible that another process had path locked and updated the readme file with new information
            readmeInfo = read_readme(readme_path)
            if readmeInfo is not None:
                currentVersion = readmeInfo['version']
                currentDate = readmeInfo['date']
                if ((currentVersion == version) and (not force)):
                    if expectedMeasuresVersion is not None:
                        # this is a release pull and the measures version must also match
                        # start off assuming a pull is necessary
                        do_pull = True
                        measures_readme_path = os.path.join(path,'geodetic/readme.txt')
                        if os.path.exists(measures_readme_path):
                            measuresReadmeInfo = read_readme(measures_readme_path)
                            if measuresReadmeInfo is not None:
                                measuresVersion = measuresReadmeInfo['version']
                                if measuresVersion == expectedMeasuresVersion:
                                    # no pull is necessary
                                    do_pull = False
                                    print_log_messages('pull_data requested "release" versions of casarundata and measures are already installed.', logger)
                            # otherwise the measures readme info is suspect and the pull should happen
                    else:
                        # nothing to do here, already at the expected version and a pull is not being forced
                        do_pull = False
                        print_log_messages('pull_data requested version is already installed.', logger)

                if do_pull:
                    # make sure the copy of installed_files is the correct one
                    installed_files = readmeInfo['extra']
                    if len(installed_files) == 0:
                        # this shoudn't happen, do not do a pull
                        do_pull = False
                        print_log_messages('destination path is not empty and the readme.txt file found there did not contain the expected list of installed files', logger, True)
                        print_log_messages('This should not happen unless multiple sessions are trying to pull_data at the same time and one experienced problems or was done out of sequence', logger, True)
                        print_log_messages('Check for other updates in process or choose a different path or clear out this path and try again', logger, True)                    
            else:
                # this shouldn't happen, do not do a pull
                do_pull = False
                print_log_messages('Unexpected error reading readme.txt file during pull_data, can not safely pull the requested version', logger, True)
                print_log_messages('This should not happen unless multiple sessions are trying to pull_data at the same time and one experienced problems or was done out of sequence', logger, True)
                
        if do_pull:
            do_pull_data(path, version, installed_files, currentVersion, currentDate, logger)
                        
        # truncate the lock file
        lock_fd.truncate(0)
        
    except Exception as exc:
        print_log_messages('ERROR! : Unexpected exception while populating casarundata version %s to %s' % (version, path), logger, True)
        print_log_messages('ERROR! : %s' % exc, logger, True)
        # leave the contents of the lock file as is to aid in debugging
        # import traceback
        # traceback.print_exc()

    # if the lock file is not closed, do that now to release the lock
    if lock_fd is not None and not lock_fd.closed:
        lock_fd.close()
        
    return
